// ======================================================================
//
// Bitmap.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "First.h"
#include "Bitmap.h"

#include "Gl.h"
#include "Iff.h"
#include "Material.h"
#include "Mouse.h"
#include "Palette.h"

#include <string.h>

// ======================================================================

Bitmap *Bitmap::firstBitmap;

// ======================================================================

Bitmap::Frame::Frame(void)
{
	init();
}

// ----------------------------------------------------------------------

Bitmap::Frame::~Frame(void)
{
	setMaterial(NULL);
}

// ----------------------------------------------------------------------

void Bitmap::Frame::reset(void)
{
	setMaterial(NULL);
	init();
}

// ----------------------------------------------------------------------

void Bitmap::Frame::init(void)
{
	material = NULL;
	width  = 0;
	height = 0;
	hotX   = 0;
	hotY   = 0;

	for (int i = 0; i < 4; ++i)
	{
		uv[i].u = CONST_REAL(0);
		uv[i].v = CONST_REAL(0);
	}
}

// ----------------------------------------------------------------------

void Bitmap::Frame::setMaterial(const Material *newMaterial)
{
	MaterialList::release(material);
	material = MaterialList::fetch(newMaterial);
}

// ----------------------------------------------------------------------

Bitmap::Frame &Bitmap::Frame::operator =(const Bitmap::Frame &newData)
{
	if (this != &newData)
	{
		setMaterial(newData.material);

		width  = newData.width;
		height = newData.height;
		hotX   = newData.hotX;
		hotY   = newData.hotY;

		for (int i = 0; i < 4; ++i)
		{
			uv[i].u = newData.uv[i].u;
			uv[i].v = newData.uv[i].v;
		}
	}

	return *this;
}

// ======================================================================

void Bitmap::debugDump(void)
{
}

// ======================================================================

Bitmap::Bitmap(void)
{
	init();
}

// ----------------------------------------------------------------------

Bitmap::Bitmap(const Material *material)
{
	Frame *f;

	DEBUG_FATAL(material->getId() != MaterialIdStatic, ("Bitmap:: not static material"));

	init();
	createFrames(1);

	f = frameData+0;

	f->material = MaterialList::fetch(material);

	f->width    = f->material->getStaticMaterial()->getWidth();
	f->height   = f->material->getStaticMaterial()->getHeight();

	f->hotX     = 0;
	f->hotY     = 0;

	f->uv[0].u  = 0;
	f->uv[0].v  = 0;

	f->uv[1].u  = 1;
	f->uv[1].v  = 0;

	f->uv[2].u  = 1;
	f->uv[2].v  = 1;

	f->uv[3].u  = 0;
	f->uv[3].v  = 1;
}

// ----------------------------------------------------------------------

Bitmap::Bitmap(const char *fn)
{
  Iff iff;

  init();

  iff.open(fn);
  load(iff);
}

// ----------------------------------------------------------------------

Bitmap::Bitmap(Iff &iff)
{
	init();
	load(iff);
}

// ----------------------------------------------------------------------

Bitmap::~Bitmap(void)
{
	Bitmap *b, *f;

	Mouse::bitmapDestroyed(this);

	// search for me in the bitmap linked list
	for (b = NULL, f = firstBitmap; f && f != this; b = f, f = f->nextBitmap)
		;

	// make sure I could find myself
	DEBUG_FATAL(f != this, ("Bitmap::~ self not found in bitmap list"));

	// remove me from the bitmap list
	if (b)
		b->nextBitmap = nextBitmap;
	else
		firstBitmap = nextBitmap;

	// null out the pointer, just trying to be safe.
	nextBitmap = NULL;

	// pitch the frame data
	delete [] frameData;
	numFrames = 0;
	frameData = NULL;
}

// ----------------------------------------------------------------------

void Bitmap::init(void)
{
	// add me to the bitmap linked list
	nextBitmap  = firstBitmap;
	firstBitmap = this;

	// init our data
	numFrames = 0;
	frameData = NULL;
}

// ----------------------------------------------------------------------

void Bitmap::load(Iff &iff)
{
	iff.seekForm(TAG_BITM);

		if (iff.seekForm(TAG_0001, Iff::FlagOptional))
		{
			load_0001(iff);
			iff.exitForm();
		}
		else
		{
			load_0000(iff);
		}

	iff.exitForm();	
}

// ----------------------------------------------------------------------

void Bitmap::load_0001(Iff &iff)
{
	Frame  *f;
	int     i, j;
	char    materialName[128];

	createFrames(iff.blockCount());
	DEBUG_FATAL(!frameData, ("Bitmap::load_0001 no frame data"));

	for (i = 0; i < numFrames; ++i)
	{
		iff.enterChunk();

		f = frameData+i;

		iff.read_string(materialName, sizeof(materialName));

		if (i == 0 || materialName[0])
			f->material = MaterialList::fetch(materialName);
		else
			f->material = MaterialList::fetch((f-1)->material);

		f->width    = iff.read_int32();
		f->height   = iff.read_int32();
		f->hotX     = iff.read_int32();
		f->hotY     = iff.read_int32();

		for (j = 0; j < 4; ++j)
		{
			f->uv[j].u = iff.read_float();
			f->uv[j].v = iff.read_float();
		}
		
		iff.exitChunk();
	}
}

// ----------------------------------------------------------------------

void Bitmap::load_0000(Iff &iff)
{
	Frame  *f;
	int     i, j;

	createFrames(iff.blockCount());
	DEBUG_FATAL(!frameData, ("Bitmap::load_0000 no frame data"));

	for (i = 0; i < numFrames; ++i)
	{
		iff.enterChunk();

		f = frameData+i;

		f->material = MaterialList::fetch(iff.read_int32());
		f->width    = iff.read_int32();
		f->height   = iff.read_int32();
		f->hotX     = iff.read_int32();
		f->hotY     = iff.read_int32();

		for (j = 0; j < 4; ++j)
		{
			f->uv[j].u = iff.read_float();
			f->uv[j].v = iff.read_float();
		}
		
		iff.exitChunk();
	}
}

// ----------------------------------------------------------------------

void Bitmap::createFrames(int cnt)
{
	// count the number of frames
	numFrames = cnt;

	// allocate and clear the frame information
	frameData = NEW Frame[numFrames];
	memset(frameData, 0, numFrames * sizeof(*frameData));
}

// ----------------------------------------------------------------------

void Bitmap::getHotspot(int *x, int *y, int frame) const
{
	DEBUG_FATAL(frame < 0 || frame >= numFrames, ("Bitmap::getHotspot frame %d of %d", frame, numFrames));

	DEBUG_FATAL(!frameData, ("Bitmap::getHotSpot no frame data"));

	if (x)
		*x = frameData[frame].hotX;

	if (y)
		*y = frameData[frame].hotY;
}

// ----------------------------------------------------------------------

void Bitmap::draw(int x, int y, const Frame &frame, const ColorVector &lighting, real a, real scale)
{
	int       i;
	TVertex   v[4];

	for (i = 0; i < 4; i++)
	{
		v[i].z = 1;
		v[i].ooz = 1;

		if (Gl::isLightingRGB())
		{
			v[i].r = lighting.getR();
			v[i].g = lighting.getG();
			v[i].b = lighting.getB();
		}
		else
		{
			v[i].r = lighting.getM();
			v[i].g = lighting.getM();
			v[i].b = lighting.getM();
		}

		v[i].a = a;
	}

	const real x0 = static_cast<real>(x - frame.hotX);
	const real y0 = static_cast<real>(y - frame.hotY);
	const real x1 = x0 + static_cast<real>(frame.width) * scale;
	const real y1 = y0 + static_cast<real>(frame.height) * scale;

	v[0].x = x0;
	v[0].y = y0;
	v[0].u = frame.uv[0].u;
	v[0].v = frame.uv[0].v;

	v[1].x = x1;
	v[1].y = y0;
	v[1].u = frame.uv[1].u;
	v[1].v = frame.uv[1].v;

	v[2].x = x1;
	v[2].y = y1;
	v[2].u = frame.uv[2].u;
	v[2].v = frame.uv[2].v;

	v[3].x = x0;
	v[3].y = y1;
	v[3].u = frame.uv[3].u;
	v[3].v = frame.uv[3].v;

	Gl::setMaterial(frame.material->getStaticMaterial());
	Gl::drawTriangleFan(v, 4);
}

// ----------------------------------------------------------------------

void Bitmap::draw(int x, int y, int frame, const ColorVector &lighting, real a) const
{
	// make sure the frame is valid
	if (frame < 0 && frame >= numFrames)
	{
		DEBUG_FATAL(true, ("Bitmap::draw frame %d of %d", frame, numFrames));
		return; //lint !e527 // unreachable
	}

	DEBUG_FATAL(!frameData, ("Bitmap::draw no frame data"));
	draw(x, y, frameData[frame], lighting, a);
}

// ----------------------------------------------------------------------

void Bitmap::write(Iff *iff) const
{
	int          i, j;
	const Frame *f;
	const char  *materialName;

	iff->appendForm(TAG_BITM);

		iff->appendForm(TAG_0001);

			for (i = 0; i < numFrames; ++i)
			{
				int32 i32;

				iff->appendChunk(TAG_FRAM);

				DEBUG_FATAL(!frameData, ("Bitmap::write no frame data"));
				f = frameData+i;

				if (i && f->material == (f-1)->material)
					materialName = "";
				else
					materialName = f->material->getName();

				iff->appendToChunk(materialName, strlen(materialName)+1);

				i32 = f->width;
				iff->appendToChunk(&i32, sizeof(i32));

				i32 = f->height;
				iff->appendToChunk(&i32, sizeof(i32));

				i32 = f->hotX;
				iff->appendToChunk(&i32, sizeof(i32));

				i32 = f->hotY;
				iff->appendToChunk(&i32, sizeof(i32));

				for (j = 0; j < 4; ++j)
				{
					float fl;

					fl = f->uv[j].u;
					iff->appendToChunk(&fl, sizeof(fl));

					fl = f->uv[j].v;
					iff->appendToChunk(&fl, sizeof(fl));
				}
			}

		iff->exitForm();	
	iff->exitForm();	
}

// ----------------------------------------------------------------------

void Bitmap::write(const char *filename) const
{
	Iff iff(10240);

	write(&iff);
	iff.write(filename);
}

// ======================================================================

